from anonypyx import generalization
from anonypyx import microaggregation
from anonypyx import models
from anonypyx import mondrian

class Anonymizer:
    def __init__(self, df, **kwargs):
        '''
        Creates a new Anonymizer instance.  
        Checks whether the given configuration options are valid.

        Parameters
        ----------
            df : pandas.DataFrame
                The pandas DataFrame to anonymize.

        Keyword Parameters
        ------------------
            sensitive_column : str
                The name of the sensitive attribute column in the data frame. Must be set when l-diversity or t-closeness are applied. (default: None)
            feature_columns : array-like
                The names of the quasi-identifier columns in the data frame. Anonymization only changes these columns. (default: All columns)
            k : int
                Parameter k of k-anonymity. (default: 1)
            l : int
                Parameter l of l-diversity. Setting this to None deactivates l-diversity. (default: None)
            t : float
                Parameter t of t-closeness. Setting this to None deactivates t-closeness. (default: None)
            diversity_definition : str
                Instantiation of l-diversity principle to use. Can be either "distinct", "simple", "entropy" or "recursive". (default: "distinct")
            closeness_metric : str
                Distance metric used by t-closeness. Can be either "max distance" or "earth mover's distance". (default: "max distance")
            algorithm : str
                The anonymization algorithm to use. Can be either "Mondrian" or "MDAV-generic" (supports only k-anonymity). (default: "Mondrian")

        Raises
        ------
            TypeError 
                When a parameter has the wrong type.
            ValueError 
                When a single parameter value or a combination of parameter values are invalid.
        
        '''
        sensitive_attribute = kwargs.get("sensitive_column", None)
        quasi_identifiers = kwargs.get("feature_columns", df.columns)
        anonymity_models = []
        k = kwargs.get("k", 1)
        l = kwargs.get("l", None)
        t = kwargs.get("t", None)
        l_diversity_definition = kwargs.get("diversity_definition", "distinct")
        t_closeness_metric = kwargs.get("closeness_metric", "max distance")
        algorithm = kwargs.get("algorithm", "Mondrian")
    
        if type(k) is not int:
            raise TypeError("k must be an integer.")
    
        if (l is not None) and (type(l) is not int):
            raise TypeError("l must be an integer.")
    
        if (t is not None) and (type(t) is not float):
            raise TypeError("t must be a float.")
    
        if type(l_diversity_definition) is not str:
            raise TypeError("diversity_definition must be a string")
        
        if type(algorithm) is not str:
            raise TypeError("algorithm must be a string")
    
        if k < 1 or k > len(df.index):
            raise ValueError("k must be between 1 and the number of records.")
    
        if (l is not None) and (l < 1):
            raise ValueError("l must be greater than 1.")
    
        # TODO: feasibility check for large l values?
    
        if (t is not None) and (t < 0 or t > 1):
            raise ValueError("t must be between 0 and 1.")
    
        if (sensitive_attribute is not None) and (sensitive_attribute not in df.columns):
            raise ValueError("sensitive_column must be a column name within the given data frame.")
    
        if not all(qi in df.columns for qi in quasi_identifiers):
            raise ValueError("Every feature column in feature_columns must be a column name within the given data frame.")
    
        privacy_models = [models.kAnonymity(k)]
    
        if l is not None:
            if l_diversity_definition == "distinct":
                privacy_models.append(models.DistinctLDiversity(l, sensitive_attribute))
            elif l_diversity_definition == "simple":
                raise NotImplementedError("Simple l-diversity has not been implemented yet.")
            elif l_diversity_definition == "entropy":
                raise NotImplementedError("Entropy l-diversity has not been implemented yet.")
            elif l_diversity_definition == "recursive":
                raise NotImplementedError("Recursive (c,l)-diversity has not been implemented yet.")
            else:
                raise ValueError("diversity_definition does not match any known instantiation of the l-diversity principle.")
    
        if t is not None:
            if t_closeness_metric == "max distance":
                privacy_models.append(models.tCloseness(t, df, sensitive_attribute, models.max_distance_metric))
            elif t_closeness_metric == "earth mover's distance":
                if not all(df[qi].dtype.name == "category" for qi in quasi_identifiers):
                    raise NotImplementedError("Earth mover's distance has not been implemented for numerical attributes yet.")
                privacy_models.append(models.tCloseness(t, df, sensitive_attribute, models.earth_movers_distance_categorical))
    
        if algorithm == "Mondrian":
            self.algorithm = mondrian.Mondrian(df, quasi_identifiers)
            self.parameters = privacy_models

            m = mondrian.Mondrian(df, quasi_identifiers)
            partitions = m.partition(privacy_models)
        elif algorithm == "MDAV-generic":
            if l is not None:
                raise ValueError("algorithm 'MDAV-generic' does not support l-diversity.")
            if t is not None:
                raise ValueError("algorithm 'MDAV-generic' does not support t-closeness.")
    
            self.algorithm = microaggregation.MDAVGeneric(df, quasi_identifiers)
            self.parameters = k
        self.df = df
        self.quasi_identifiers = quasi_identifiers
        self.sensitive_attribute = sensitive_attribute

    def anonymize(self):
        '''
        Starts the anonymization algorithm with the options specified by this Anonymizer instance.

        Returns
        -------
            List of anonymized records.
        '''
        partitions = self.algorithm.partition(self.parameters)
        return generalization.aggregate_partitions(self.df, partitions, self.quasi_identifiers, self.sensitive_attribute)




import secrets
import pandas as pd
import numpy as np
from scipy.spatial.distance import cdist, pdist, squareform

def categorical_distance(row1, row2):
    return np.sum(row1 != row2)

class MDAVGeneric:
    '''
    Implements the MDAV-generic algorithm from [1].

    TODO: maybe add improvements from [2]?

    [1]: Domingo-Ferrer, J., & Torra, V. (2005). Ordinal, continuous and heterogeneous 
    k-anonymity through microaggregation. Data Mining and Knowledge Discovery, 11, 195–212.

    [2]: Rodríguez-Hoyos, A., Estrada-Jiménez, J., Rebollo-Monedero, D., Mezher, A. M., 
    Parra-Arnau, J., & Forné, J. (2020). The Fast Maximum Distance to Average Vector 
    (F-MDAV): An algorithm for k-anonymous microaggregation in big data. Engineering 
    Applications of Artificial Intelligence, 90, 103531. 
    '''
    def __init__(self, df, feature_columns):
        self.df = df
        self.feature_columns = feature_columns
        self.num_data_points = len(self.df.index)

    def partition(self, k):
        self.__prepare_data()
        self.__build_distance_matrix()
        self.clusters = []

        while (len(self.remaining_indices) >= 3 * k):
            centroid = self.__find_centroid()
            far_end_dist = self.__get_distance_vector_of_most_distant_point(centroid)
            self.__assign_closest_points_to_new_cluster(far_end_dist, k)

            other_end_dist = self.__get_distance_vector_of_most_distant_point_from_data_point(far_end_dist.loc[self.remaining_indices])
            self.__assign_closest_points_to_new_cluster(other_end_dist, k)

        if len(self.remaining_indices) >= 2 * k:
            centroid = self.__find_centroid()
            far_end_dist = self.__get_distance_vector_of_most_distant_point(centroid)
            self.__assign_closest_points_to_new_cluster(far_end_dist, k)

        self.clusters.append(self.remaining_indices)

        return self.clusters 

    def __prepare_data(self):
        self.remaining = self.df.drop(columns=[column for column in self.df.columns if column not in self.feature_columns])
        self.remaining_indices = self.remaining.index
        self.categorical = [column for column in self.feature_columns if self.remaining[column].dtype == "category"]
        self.continuous  = [column for column in self.feature_columns if self.remaining[column].dtype != "category"]
       
        self.remaining.loc[:, self.continuous] = (self.remaining.loc[:, self.continuous] - self.remaining.loc[:, self.continuous].mean()) / self.remaining.loc[:, self.continuous].std()

        for attribute in self.categorical:
            # self.categorical.loc[:, attribute] = pd.factorize(self.categorical.loc[:, attribute])[0]
            self.remaining.loc[:, attribute] = pd.factorize(self.remaining.loc[:, attribute])[0]

    def __find_centroid(self):
        centroid_continuous = self.remaining.loc[self.remaining_indices, self.continuous].mean()
        modes = self.remaining.loc[self.remaining_indices, self.categorical].mode()
        centroid_categorical = [secrets.choice(modes.loc[:, column].dropna()) for column in modes]
        centroid_categorical = np.array(centroid_categorical)
        return (centroid_continuous, centroid_categorical)

    def __get_distance_vector_of_most_distant_point(self, position):
        pos_continuous, pos_categorical = position
        distances1 = cdist([pos_continuous], self.remaining.loc[self.remaining_indices, self.continuous], metric='euclidean')[0]
        distances2 = cdist([pos_categorical], self.remaining.loc[self.remaining_indices, self.categorical], metric='hamming')[0] * len(self.categorical)
        distances = distances1 + distances2
        max_dist_array_pos = distances.argmax()
        max_dist_index = self.remaining_indices[max_dist_array_pos]
        return self.distance_matrix.loc[max_dist_index, self.remaining_indices]

    def __get_distance_vector_of_most_distant_point_from_data_point(self, distance_vector):
        max_dist_array_pos = distance_vector.argmax()
        max_dist_index = self.remaining_indices[max_dist_array_pos]
        return self.distance_matrix.loc[max_dist_index, self.remaining_indices]

    def __assign_closest_points_to_new_cluster(self, distance_vector, k):
        array_positions = np.argpartition(distance_vector, k)[:k]

        indices = self.remaining_indices.take(array_positions)
        self.remaining_indices = self.remaining_indices.drop(indices)

        self.clusters.append(indices)

    def __build_distance_matrix(self):
        dist_continuous = pdist(self.remaining.loc[:, self.continuous], metric='euclidean')
        dist_categorical = (pdist(self.remaining.loc[:, self.categorical], metric='hamming') * len(self.categorical)) if not len(self.categorical) == 0 else 0

        self.distance_matrix = pd.DataFrame(squareform(dist_categorical + dist_continuous))




def agg_categorical_column(series):
    # this is workaround for dtype bug of series
    series.astype("category")

    # l = [str(n) for n in set(series)]
    l = [str(n) for n in series.unique()]
    return ",".join(l)


def agg_numerical_column(series):
    minimum = series.min()
    maximum = series.max()
    if maximum == minimum:
        string = str(maximum)
    else:
        string = f"{minimum}-{maximum}"
    return string


def aggregate_partitions(df, partitions, feature_columns, sensitive_column):
    # TODO: inplace generalization to increase performance?
    aggregations = {}
    for column in feature_columns:
        if df[column].dtype.name == "category":
            aggregations[column] = agg_categorical_column
        else:
            aggregations[column] = agg_numerical_column
    rows = []
    for i, partition in enumerate(partitions):
        rows += generalize_partition(df, partition, aggregations, sensitive_column)
    return rows

def count_sensitive_values_in_partition(df, partition, sensitive_column):
    return df.loc[partition].groupby(sensitive_column, observed=True).agg({sensitive_column: "count"})

def generalize_partition(df, partition, aggregations, sensitive_column):
    rows = []
    # partition -> generalized quasi-identifier + sensitive value distribution
    values = df.loc[partition].agg(aggregations, squeeze=False)
    sensitive_counts = count_sensitive_values_in_partition(df, partition, sensitive_column)
    for sensitive_value, count in sensitive_counts[sensitive_column].items():
        # if count == 0:
        #     continue
        values[sensitive_column] = sensitive_value
        values["count"] = count
        rows.append(values.copy())
    return rows



class Mondrian:
    def __init__(self, df, feature_columns):
        self.df = df
        self.feature_columns = feature_columns

    def get_spans(self, partition, scale=None):
        spans = {}
        for column in self.feature_columns:
            if self.df[column].dtype.name == "category":
                span = len(self.df[column][partition].unique())
            else:
                span = (
                    self.df[column][partition].max() - self.df[column][partition].min()
                )
            if scale is not None:
                span = span / scale[column]
            spans[column] = span
        return spans

    def split(self, column, partition):
        dfp = self.df[column][partition]
        if dfp.dtype.name == "category":
            values = dfp.unique()
            lv = set(values[: len(values) // 2])
            rv = set(values[len(values) // 2 :])
            return dfp.index[dfp.isin(lv)], dfp.index[dfp.isin(rv)]
        else:
            median = dfp.median()
            dfl = dfp.index[dfp < median]
            dfr = dfp.index[dfp >= median]
            return (dfl, dfr)

    def partition(self, privacy_models):
        scale = self.get_spans(self.df.index)

        finished_partitions = []
        partitions = [self.df.index]

        while partitions:
            partition = partitions.pop(0)
            normalized_spans = self.get_spans(partition, scale)
            for column, span in sorted(normalized_spans.items(), key=lambda x: -x[1]):
                left_part, right_part = self.split(column, partition)

                if self.__all_models_enforceable(left_part, right_part, privacy_models):
                    partitions.extend((left_part, right_part))
                    break
            else:
                finished_partitions.append(partition)
        return finished_partitions

    def __all_models_enforceable(self, left_part, right_part, privacy_models):
        return all(model.is_enforcable(self.df.loc[left_part]) and model.is_enforcable(self.df.loc[right_part]) for model in privacy_models)




import anonypyx
from anonypyx import models
from anonypyx import mondrian
import pandas as pd

data = [
    [6, "1", "test1", "x", 20],
    [6, "1", "test1", "x", 30],
    [8, "2", "test2", "x", 50],
    [8, "2", "test3", "w", 45],
    [8, "1", "test2", "y", 35],
    [4, "2", "test3", "y", 20],
    [4, "1", "test3", "y", 20],
    [2, "1", "test3", "z", 22],
    [2, "2", "test3", "y", 32],
]

columns = ["col1", "col2", "col3", "col4", "col5"]
categorical = set(("col2", "col3", "col4"))

def test_k_anonymity():
    df = pd.DataFrame(data=data, columns=columns)

    for name in categorical:
        df[name] = df[name].astype("category")

    feature_columns = ["col1", "col2", "col3"]
    m = mondrian.Mondrian(df, feature_columns)
    partitions = m.partition([models.kAnonymity(2)])
    print(f"partitions: {partitions}")


def test_distinct_l_diversity():
    df = pd.DataFrame(data=data, columns=columns)

    for name in categorical:
        df[name] = df[name].astype("category")

    feature_columns = ["col1", "col2", "col3"]
    sensitive_column = "col4"

    m = mondrian.Mondrian(df, feature_columns)
    partitions = m.partition([models.DistinctLDiversity(2, sensitive_column)])

    print(f"partitions: {partitions}")


def test_t_closeness():
    df = pd.DataFrame(data=data, columns=columns)

    for name in categorical:
        df[name] = df[name].astype("category")

    feature_columns = ["col1", "col2", "col3"]
    sensitive_column = "col4"

    m = mondrian.Mondrian(df, feature_columns)
    partitions = m.partition([models.tCloseness(0.2, df, sensitive_column, models.max_distance_metric)])

    print(f"partitions: {partitions}")


def test_get_spans():
    df = pd.DataFrame(data=data, columns=columns)

    for name in categorical:
        df[name] = df[name].astype("category")

    feature_columns = ["col1", "col2", "col3"]

    m = mondrian.Mondrian(df, feature_columns)
    spans = m.get_spans(df.index)

    assert {"col1": 6, "col2": 2, "col3": 3} == spans

def test_get_spans_with_scale():
    df = pd.DataFrame(data=data, columns=columns)
    scale = {"col1": 6, "col2": 4, "col3": 5}

    for name in categorical:
        df[name] = df[name].astype("category")

    feature_columns = ["col1", "col2", "col3"]

    m = mondrian.Mondrian(df, feature_columns)
    spans = m.get_spans(df.index, scale)

    assert {"col1": 6/6, "col2": 2/4, "col3": 3/5} == spans



from anonypyx import generalization
import pandas as pd

def test_agg_categorical_column():
    series = pd.Series(['A','B'])
    result = generalization.agg_categorical_column(series) 
    assert (result == "A,B" or result == "B,A")


def test_agg_numerical_column():
    series = pd.Series([10,6,7,7,6,4,2,3])
    assert generalization.agg_numerical_column(series) == "2-10"


def test_count_sensitive_values_in_partition():
    df = pd.DataFrame({"QI1": [101,102,103,110,110], "QI2": ["A","A","B","B","C"], "S":[10,20,10,21,10]}, index=[1,2,3,4,5])
    partition = [1,2,3]
    sensitive_column = "S"
    result = generalization.count_sensitive_values_in_partition(df, partition, sensitive_column) 
    assert result.to_dict() == {'S': {10: 2, 20: 1}}


def test_generalize_partition():
    df = pd.DataFrame({"QI1": [101,102,103,110,110], "QI2": [5,5,4,4,8], "S":[10,20,10,21,10]}, index=[1,2,3,4,5])
    partition = [1,2,3]
    sensitive_column = "S"
    aggregations = {"QI1": generalization.agg_numerical_column, "QI2": generalization.agg_numerical_column}
    result = generalization.generalize_partition(df, partition, aggregations, sensitive_column)
    result = pd.DataFrame(result)
    assert result.equals(pd.DataFrame([{'QI1': '101-103', 'QI2': '4-5', 'S': 10, 'count': 2}, {'QI1': '101-103', 'QI2': '4-5', 'S': 20, 'count': 1}]))



import anonypyx
import pandas as pd

import pytest

@pytest.fixture
def prepared_df():
    data = [
        [6, "1", "test1", "x", 20],
        [6, "1", "test1", "x", 30],
        [8, "2", "test2", "x", 50],
        [8, "2", "test3", "w", 45],
        [8, "1", "test2", "y", 35],
        [4, "2", "test3", "y", 20],
        [4, "1", "test3", "y", 20],
        [2, "1", "test3", "z", 22],
        [2, "2", "test3", "y", 32],
    ]
    df = pd.DataFrame(data=data, columns=["col1", "col2", "col3", "col4", "col5"])
    for name in ("col2", "col3", "col4"):
        df[name] = df[name].astype("category")

    return df

def test_k_anonymity(prepared_df):
    a = anonypyx.Anonymizer(prepared_df, k=2, feature_columns=["col1", "col2", "col3"], sensitive_column="col4")
    rows = a.anonymize()

    dfn = pd.DataFrame(rows)
    print(dfn)


def test_l_diversity(prepared_df):
    a = anonypyx.Anonymizer(prepared_df, k=2, l=2, diversity_definition="distinct", feature_columns=["col1", "col2", "col3"], sensitive_column="col4")
    rows = a.anonymize()

    dfn = pd.DataFrame(rows)
    print(dfn)


def test_t_closeness(prepared_df):
    a = anonypyx.Anonymizer(prepared_df, k=2, t=0.2, closeness_metric="max distance", feature_columns=["col1", "col2", "col3"], sensitive_column="col4")
    rows = a.anonymize()

    dfn = pd.DataFrame(rows)
    print(dfn)




import anonypyx
from anonypyx import models

import pandas as pd
import pytest

@pytest.fixture
def k_anonymity_df():
    data = [
        [1, "A", 5],
        [1, "A", 6],
        [1, "A", 7]
    ]
    df = pd.DataFrame(data=data, columns=["col1", "col2", "col3"])
    df["col2"] = df["col2"].astype("category")

    return df

@pytest.fixture
def distinct_l_diversity_df():
    data = [
        [1, "A", 5],
        [1, "A", 5],
        [1, "A", 6],
        [1, "A", 6],
        [1, "A", 7]
    ]
    df = pd.DataFrame(data=data, columns=["col1", "col2", "col3"])
    df["col2"] = df["col2"].astype("category")

    return df

@pytest.fixture
def t_closeness_df():
    parent_data = [
        ["1", "A", "1"],
        ["1", "B", "1"],
        ["1", "A", "2"],
        ["1", "B", "3"],
        ["1", "A", "3"],
        ["1", "B", "4"],
    ]
    parent_df = pd.DataFrame(data=parent_data, columns=["col1", "col2", "col3"])
    parent_df["col2"] = parent_df["col2"].astype("category")
    parent_df["col3"] = parent_df["col3"].astype("category")
    
    return parent_df

def test_max_distance_metric():
    dist1 = {"A": 0.2, "B": 0.4, "C": 0.6}
    dist2 = {"A": 0.05, "B": 0.3, "C": 0.65}

    assert pytest.approx(0.15, 0.001) == models.max_distance_metric(dist1, dist2)

def test_k_anonymity(k_anonymity_df):
    sensitive_column = "col3"
    assert models.kAnonymity(3).is_enforcable(k_anonymity_df)

def test_k_anonymity_detects_violation(k_anonymity_df):
    sensitive_column = "col3"
    assert not (models.kAnonymity(4).is_enforcable(k_anonymity_df))

def test_distinct_l_diversity(distinct_l_diversity_df):
    sensitive_column = "col3"
    assert (models.DistinctLDiversity(3, sensitive_column).is_enforcable(distinct_l_diversity_df))

def test_distinct_l_diversity_detects_violation(distinct_l_diversity_df):
    sensitive_column = "col3"
    assert not (models.DistinctLDiversity(4, sensitive_column).is_enforcable(distinct_l_diversity_df))

def test_maxdistance_t_closeness(t_closeness_df):
    df = t_closeness_df.loc[[0, 2]] 
    sensitive_column = "col3"
    assert (models.tCloseness(2.01/6.0, t_closeness_df, sensitive_column, models.max_distance_metric).is_enforcable(df))

def test_maxdistance_t_closeness_detects_violation(t_closeness_df):
    df = t_closeness_df.loc[[0, 2]] 
    sensitive_column = "col3"
    assert not (models.tCloseness(1.99/6.0, t_closeness_df, sensitive_column, models.max_distance_metric).is_enforcable(df))

def test_emd_t_closeness_categorical(t_closeness_df):
    df = t_closeness_df.loc[[0, 2]] 
    sensitive_column = "col3"
    assert (models.tCloseness(0.5, t_closeness_df, sensitive_column, models.earth_movers_distance_categorical).is_enforcable(df))

def test_emd_t_closeness_detects_violation_categorical(t_closeness_df):
    df = t_closeness_df.loc[[0, 2]] 
    sensitive_column = "col3"
    assert not (models.tCloseness(0.49, t_closeness_df, sensitive_column, models.earth_movers_distance_categorical).is_enforcable(df))

def test_k_anonymity_no_sensitive_attribute(k_anonymity_df):
    sensitive_column = None
    assert models.kAnonymity(3).is_enforcable(k_anonymity_df)

def test_l_diverse_no_sensitive_attribute(distinct_l_diversity_df):
    sensitive_column = None
    assert not (models.DistinctLDiversity(3, sensitive_column).is_enforcable(distinct_l_diversity_df))

def test_t_closeness_no_sensitive_attribute(t_closeness_df):
    df = t_closeness_df.loc[[0, 2]] 
    sensitive_column = None
    assert not (models.tCloseness(0.5, t_closeness_df, sensitive_column, models.earth_movers_distance_categorical).is_enforcable(df))



class kAnonymity:
    def __init__(self, k): 
        self.__k = k

    def is_enforcable(self, df):
        return len(df.index) >= self.__k

class DistinctLDiversity:
    def __init__(self, l, sensitive_column):
        self.__l = l
        self.__sensitive_column = sensitive_column

    def is_enforcable(self, df):
        if self.__sensitive_column is None:
            return False
        return self.__l <= len(df[self.__sensitive_column].unique())

def earth_movers_distance_categorical(distribution1, distribution2):
    diff_sum = 0.0

    for value, f1 in distribution1.items():
        f2 = distribution2[value]
        diff_sum += abs(f1 - f2)

    return diff_sum * 0.5

def max_distance_metric(distribution1, distribution2):
    max_diff = 0

    for value, f1 in distribution1.items():
        f2 = distribution2[value]
        max_diff = max(abs(f1 - f2), max_diff)

    return max_diff

class tCloseness:
    def __init__(self, t, df, sensitive_column, distance_metric):
        self.__t = t
        self.__sensitive_column = sensitive_column
        if self.__sensitive_column is not None:
            self.__global_frequencies = get_frequency(df, sensitive_column)
        self.__metric = distance_metric

    def is_enforcable(self, df):
        if self.__sensitive_column is None:
            return False
        total_count = float(len(df.index))

        if total_count == 0:
            return False

        # TODO: optimization possible: do not recalculate frequencies
        local_frequencies = get_frequency(df, self.__sensitive_column)

        return self.__metric(local_frequencies, self.__global_frequencies) <= self.__t

def get_frequency(df, sensitive_column):
    global_freqs = {}
    total_count = float(len(df))
    group_counts = df.groupby(sensitive_column, observed=False)[sensitive_column].agg("count")

    for value, count in group_counts.to_dict().items():
        p = count / total_count
        global_freqs[value] = p
    return global_freqs



import pandas as pd
import numpy as np
from scipy.spatial.distance import pdist, squareform
from anonypyx.microaggregation import MDAVGeneric

def test_MDAVGeneric_continuous_data_already_normalized_for_debugging():
    data = [
        [ 1.5,  1.5,  1.5],    # farthest from mean at (0,0,0)
        [ 0.5,  0.5,  0.5],    # cloest to this point
        [-1.5, -1.5, -0.5], # farthest from (1.5, 1.5, 1.5)
        [   0, -0.5, -1.5],    # closest to (-1.5, -1.5, -0.5)
        [   0,    0,    0],          # remaining
        [ 0.5,    0,    0],        # remaining
    ]
    columns = ["num1", "num2", "num3"]
    df = pd.DataFrame(data=data, columns=columns)

    partitioner = MDAVGeneric(df, columns)

    expected = [list(df.index[i:i+2]) for i in range(0, 6, 2)]
    actual = partitioner.partition(2)
    # sort to make order deterministic
    sorted_actual = sorted([sorted(subset) for subset in actual])

    assert expected == sorted_actual

def test_MDAVGeneric_continuous_data():
    data = [
        [0, 0, 0],
        [0, 1, -1],
        [100, 99.9, 101.1],
        [100, 100, 100],
        [0, 200, 0],
        [0, 200, 0],
        [-500, 3, 2],
        [-510, 3.141, 2],
    ]
    columns = ["num1", "num2", "num3"]
    df = pd.DataFrame(data=data, columns=columns)

    partitioner = MDAVGeneric(df, columns)

    expected = [list(df.index[i:i+2]) for i in range(0, 8, 2)]
    actual = partitioner.partition(2)
    # sort to make order deterministic
    sorted_actual = sorted([sorted(subset) for subset in actual])

    assert expected == sorted_actual

def test_MDAVGeneric_categorical_data():
    data = [
        ["A", "1", "foo", "mode", "a"],
        ["A", "1", "foo", "mode", "b"],
        ["A", "1", "bar", "mode", "c"],
        ["A", "1", "bar", "mode", "d"],
        ["B", "2", "bar", "no mode", "e"],
        ["B", "2", "bar", "no mode", "f"],
    ]
    
    columns = ["col1", "col2", "col3", "col4", "col5"]
    categorical = set(columns)

    df = pd.DataFrame(data=data, columns=columns)

    for name in categorical:
        df[name] = df[name].astype("category")

    partitioner = MDAVGeneric(df, columns)

    expected = [list(df.index[i:i+2]) for i in range(0, 6, 2)]
    actual = partitioner.partition(2)
    # sort to make order deterministic
    sorted_actual = sorted([sorted(subset) for subset in actual])

    assert expected == sorted_actual

def test_MDAVGeneric_mixed_data():
    data = [
        [0, 0, 0, "A"],
        [0, 1, -1, "A"],
        [50, 49.9, 101.1, "B"],
        [50, 50, 100, "B"],
        [0, 200, 0, "C"],
        [0, 200, 0, "C"],
        [-500, 3, 2, "C"],
        [-510, 3.141, 2, "C"],
    ]
    columns = ["num1", "num2", "num3", "cat4"]
    categorical = set(["cat4"])

    df = pd.DataFrame(data=data, columns=columns)

    for name in categorical:
        df[name] = df[name].astype("category")

    partitioner = MDAVGeneric(df, columns)

    expected = [list(df.index[i:i+2]) for i in range(0, 8, 2)]
    actual = partitioner.partition(2)
    # sort to make order deterministic
    sorted_actual = sorted([sorted(subset) for subset in actual])

    assert expected == sorted_actual

def test_MDAVGeneric_excluding_columns_works():
    data = [
        [0, 0, 0],
        [0, 1, -1],
        [100, 99.9, 101.1],
        [100, 100, 100],
        [0, 200, 0],
        [0, 200, 0],
    ]
    columns = ["num1", "num2", "num3"]
    df = pd.DataFrame(data=data, columns=columns)

    partitioner = MDAVGeneric(df, ["num1", "num2"])

    expected = [list(df.index[i:i+2]) for i in range(0, 6, 2)]
    actual = partitioner.partition(2)
    # sort to make order deterministic
    sorted_actual = sorted([sorted(subset) for subset in actual])

    assert expected == sorted_actual

def test_MDAVGeneric_does_not_alter_data():
    data = [
        [0, 0, 0],
        [0, 1, -1],
        [100, 99.9, 101.1],
        [100, 100, 100],
        [0, 200, 0],
        [0, 200, 0],
    ]
    columns = ["num1", "num2", "num3"]
    df = pd.DataFrame(data=data, columns=columns)

    expected = df.copy()

    partitioner = MDAVGeneric(df, ["num1", "num2"])
    partitioner.partition(2)

    assert df.equals(expected)

